/* paddr.c 
 * 
 * Antonio Barbalace, SSRG Virignia Tech, 2015
 * 
 * parts of this application derive from the Linux kernel
 * specifically Documentation/vm/page-types.c
 * 
 * TODO make sure that the entries are directed flushed to the
 * HD so that we are not consuming to much memory footprint
 * 
 * NOTE 
 * flags 0x400 is buddy (count -127 ?) cat paddr.out | grep 0x00000400 | awk '{ if ($4 != "-127") print $0 }' | wc -l
 RETURNS 0
 * flags 0x800 is mmap (first 16 pages) - 64kB
 * flags 0x1 0000 0000 ?!
 * flags 0x4 0000 086c ?!
 * flags 0x100000 is nopage - basically PCI hole
 *
 */

#define _GNU_SOURCE

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdint.h>
#include <limits.h>
#include <sys/types.h>
#include <sys/fcntl.h>
#include <sys/statfs.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "paddr.h"

static int page_size;

static int kpagecount_fd;
static int kpageflags_fd;
static int kpagepaddr_fd;

static int my_open(const char*  pathname, int flags)
{
        int fd = open(pathname, flags);
        if (fd < 0) {
                perror(pathname);
                exit(1);
        }
        return fd;
}

static unsigned long do_u64_read(int fd, char *name, uint64_t *buf, unsigned long index, unsigned count) 
{
        long bytes;
        if (index > ULONG_MAX/8) {
                dprintf(STDERR_FILENO, "limit error\n");
                exit(1);
        }
        if (lseek(fd, index*8, SEEK_SET)<0) {
                perror(name);
                exit(1);
        }
        bytes = read(fd, buf, count*8);
        if ( bytes < 0) {
                perror(name);
                exit(1);
        }
        if (bytes %8) {
                dprintf(STDERR_FILENO, "alingment error\n");
                exit(1);
        }
        return bytes/8;
}

#define PROC_KPAGECOUNT "/proc/kpagecount"
#define PROC_KPAGEFLAGS "/proc/kpageflags"
#define PROC_KPAGEPADDR "/proc/kpagepaddr"

static inline unsigned long kpagecount_read(uint64_t *buf, unsigned long index, unsigned long pages)
{
	return do_u64_read(kpagecount_fd, PROC_KPAGECOUNT, buf, index, pages);
}

static inline unsigned long kpageflags_read(uint64_t *buf, unsigned long index, unsigned long pages)
{
	return do_u64_read(kpageflags_fd, PROC_KPAGEFLAGS, buf, index, pages);
}

static inline unsigned long kpagepaddr_read(uint64_t *buf, unsigned long index, unsigned long pages)
{
	return do_u64_read(kpagepaddr_fd, PROC_KPAGEPADDR, buf, index, pages);
}

#define min_t(type, x , y) ({ type __min1 = (x); type __min2 = (y); __min1 < __min2 ? __min1 : __min2; })

unsigned long paddr = -1UL;

int ofd = 0;
void* the_bufferedoutput = 0;
int the_bufferedindex = 0;
char * output_file =0;

#define KPAGE_BATCH (64 << 10)
static void walk_pfn(unsigned long voffset, unsigned long index, unsigned long count)
{
        uint64_t *buf_paddr, *buf_count, *buf_flags;
        unsigned long batch;
        long pages_paddr, pages_count, pages_flags;
        unsigned long i;
	
	buf_paddr = malloc (KPAGE_BATCH * sizeof(uint64_t));
	buf_count = malloc (KPAGE_BATCH * sizeof(uint64_t));
	buf_flags = malloc (KPAGE_BATCH * sizeof(uint64_t));
	if (!buf_paddr || !buf_count || !buf_flags) {
	  dprintf(STDERR_FILENO, "malloc error buf _paddr %p _count %p _flags %p\n",
	    buf_paddr, buf_count, buf_flags );
	  exit(1);
	}

        while (count) {
                batch = min_t(unsigned long, count, KPAGE_BATCH);
                pages_paddr = kpagepaddr_read(buf_paddr, index, batch);
		pages_count = kpagecount_read(buf_count, index, batch);
		pages_flags = kpageflags_read(buf_flags, index, batch);		
		
		if ((pages_paddr != pages_count) || (pages_paddr != pages_flags)) {
		  dprintf(STDERR_FILENO,"WARN cannot read same amount of pages _paddr %ld _count %ld _flags %ld\n",
		    pages_paddr, pages_count, pages_flags);
		  break;
		}
		
                if (pages_paddr == 0 || pages_count == 0 || pages_flags == 0)
                        break;

                for (i = 0; i < pages_paddr; i++) {
// TODO check the input switch
// TODO add error checking if possible
                        if ( 0 ) {
                                printf("%lx %lx 0x%08lx %ld 0x%08lx\n",
				       voffset + i, index + i, 
				       buf_paddr[i], buf_count[i], buf_flags[i]);
			}
			else {
				paddr_desc pdesc;
				unsigned long * buffer = (unsigned long *) &pdesc; //unsigned long buffer[4];
				/*buffer[0] = (index+i)*page_size ;
				buffer[1] = buf_paddr[i];
				buffer[2] = buf_count[i];
				buffer[3] = buf_flags[i];*/
				pdesc.pfn = (index+i)*page_size ;
				pdesc.phys = buf_paddr[i];
				pdesc.count = buf_count[i];
				pdesc.flags = buf_flags[i];

				if ( !output_file ) {
					ssize_t ret = write(STDOUT_FILENO, buffer, (4 * sizeof(unsigned long)));
					if (ret < 0) {
					  perror("ERROR while writing on stdout");
					  exit(1);
					}
				}
				else {
					int size = sizeof(pdesc); //(4 *sizeof(unsigned long));
					
					if ( (the_bufferedindex + size) < page_size ) {
						// only write in the buffer
						memcpy( (the_bufferedoutput + the_bufferedindex), buffer, size);
						the_bufferedindex += size;
					}
					else {
						if ( (page_size - the_bufferedindex) > 0 ) {
							memcpy( (the_bufferedoutput + the_bufferedindex), buffer, (page_size - the_bufferedindex));
							the_bufferedindex = page_size;
							size -= (page_size - the_bufferedindex);
						}
						
						ssize_t ret = write(ofd, the_bufferedoutput, the_bufferedindex);
						if (ret < 0) {
							perror("ERROR while writing on stdout");
							exit(1);
						}
						fdatasync(ofd);
						the_bufferedindex = 0;
						  
						if (size > 0) {
							memcpy(the_bufferedoutput, ((char*)buffer) + (sizeof(pdesc) - size), size);
							the_bufferedindex +=size;
						}
					}				  
				}
			}
/*
                        if (paddr == -1UL) {
                                paddr = buf[i];
                                printf("this should be single\n"); // remove
                                printf("0x%lx-", paddr);
                        }
                        else {
                                if ((paddr + 0x40) != buf[i]) {
                                        printf("0x%lx\n", paddr);
                                        printf("0x%lx-", buf[i]);
                                }
                                paddr = buf[i];
                        }
                        */
                }
                index += pages_paddr;
		voffset += pages_paddr;
                count -= pages_paddr;
        }
        
        if (the_bufferedindex) {
		ssize_t ret = write(ofd, the_bufferedoutput, the_bufferedindex);
		if (ret < 0) {
			perror("ERROR while writing on stdout");
			exit(1);
		}
		fdatasync(ofd);
	}
        
        free (buf_paddr);
	free (buf_count);
	free (buf_flags);
}

static void walk_addr_ranges(void)
{
        int i;
	kpagecount_fd = my_open(PROC_KPAGECOUNT, O_RDONLY);
	kpageflags_fd = my_open(PROC_KPAGEFLAGS, O_RDONLY);
        kpagepaddr_fd = my_open(PROC_KPAGEPADDR, O_RDONLY);

        walk_pfn(0, 0, ULONG_MAX);

        close(kpagecount_fd);
	close(kpageflags_fd);
	close(kpagepaddr_fd);
}

/*
 * here we try to avoid the buffer cache (we don't want to perturbate the mapping,
 * moreover we don't care about the performance of the dump, we just want it "realistic"
 */
void* open_appbufferedoutput(char * file_path)
{
  int oflags;
  char * the_buffer;
//  struct dioattr dio; IRIX code for fcntl - not supported in Linux
  
  oflags = O_WRONLY | O_CREAT | O_DIRECT; //  O_TRUNC | 
  ofd = open(file_path, oflags, 0644);
  if (ofd < 0) {
    perror ("ERROR open app buffered output");
    exit(1);
  }
  
  int ret = posix_memalign((void**)&the_buffer, page_size, page_size );
  if (ret != 0) {
    dprintf(STDERR_FILENO, "ERROR from posix_memalign is %d\n", ret);
    exit (1);
  }
  memset(the_buffer, 0x00, page_size);
  
  return the_buffer;
}

int main(int argc, char* argv[])
{
	dprintf(STDERR_FILENO, "This application merges /proc/kpagecount, /proc/kpageflags\n" );
	dprintf(STDERR_FILENO, "and the new introduced /proc/kpagepaddr into a 8-byte fourplets\n" );
	dprintf(STDERR_FILENO, "use -v to print on stdout, no argument fourplets on stdout\n" );
	dprintf(STDERR_FILENO, "use -o output.file fourplets on the file synch (not pagebuffer)\n" );
	dprintf(STDERR_FILENO, "THE SUGGESTED WAY OF USAGE IS THE LAST ONE (file synch)\n" );

	// TODO parse arguments here
	//output_file = "paddr.dat"; // simply comment to to go on stdout

	page_size= getpagesize(); // we are assuming, as documents online that the bock size is a page
  
	if (output_file)
	  the_bufferedoutput = open_appbufferedoutput(output_file);
	
        walk_addr_ranges();
	
	if (output_file)
	  close(ofd);
	
        return 1;
}
