
 
// Antonio Barbalace, SSRG Virginia Tech, 2015

// additional code to test if we are touching the page table

#include <string.h>
#include <stdio.h>
#include <stdint.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/fcntl.h>
#include <sys/statfs.h>
#include <sys/stat.h>

typedef unsigned long u64;
 
#include "paddr.h"
//#include "linux-config.h"
//#include "memory-failure.h"
#include "page-types.h"

static int page_size;

static long elems_array = 0;
static struct page * pages_array = 0;

//#define KERN_MASK 0xffff ea00 0000 0000
#define STRUCT_PAGE 0x40
 
unsigned long group_first = (~0);
unsigned long group_last = (~0);

unsigned long physical_first = (~0);
unsigned long physical_last = (~0);

unsigned long load_array(char * input_file, long elements, struct page * parray)
{
  long l, bytes;
  long pfn_prev = -1, pfn_cur =0;
  int pfd;
  paddr_desc pdesc;
  
  pfd= open(input_file,O_RDONLY);
  if (pfd < 0) {
    perror(input_file);
    exit(1);
  }
  
  for (l=0; l<elements; l++) {
    bytes = read (pfd, (void *)&pdesc, sizeof(pdesc));
    if ( bytes < 0) {
      dprintf(STDERR_FILENO, "ERROR in reading the file %s\n", input_file);
      return -1;
    }
    if ( bytes % sizeof(pdesc) ) {
      dprintf(STDERR_FILENO, "ERROR in the alignement, read returned %ld (%ld) for %s\n", bytes, sizeof(pdesc), input_file);
      return -1;
    }
    
//    printf("%lx %lx 0x%08lx %ld 0x%08lx (%d)\n",
	   //pdesc.pfn, (pdesc.pfn /page_size), (pdesc.phys), pdesc.count, pdesc.flags, bytes);

    //check for inconsistencies (I know this is evil ...)
    pfn_cur = pdesc.pfn / page_size;
    if ((pfn_prev +1) != pfn_cur) {// we strictly assume that pfn should be contiguous
      dprintf(STDERR_FILENO, "ERROR missing pfn, pfn are not contiguous pfn %ld prev %ld\n", pfn_cur, pfn_prev); 
      return -1;
    }
    if (pfn_cur != l) {
      dprintf(STDERR_FILENO, "ERROR something wrong, pfn are not contiguous pfn %ld counter %ld\n", pfn_cur, l); 
      return -1;
    }      
    
    if (pdesc.phys == 0 && (pdesc.flags & (1 << KPF_NOPAGE)) ) {
      if (physical_first !=(~0) && physical_last != (~0)) {
	printf("physical %lx last %lx\n", physical_first, physical_last + (page_size -1));
	physical_first = (~0);
	physical_last = (~0);
      }
      goto next;
    } else {
	if (physical_first == (~0))
	    physical_first = (l*page_size);
	else
	    physical_last = (l*page_size);
    }
    
    //pdesc.phys &= ~KERN_MASK;
    
    if (group_first == (~0))
      group_first = group_last = pdesc.phys;
    else
      if ( (group_last + STRUCT_PAGE) == pdesc.phys )
	group_last = pdesc.phys;
      else {
	printf("table %lx - %lx\n", group_first, group_last);
	group_first = group_last = pdesc.phys;
      }
    
next:    
    pfn_prev=pfn_cur;
  }
  
  printf("table %lx - %lx\n", group_first, group_last);
  printf("physical %lx last %lx\n", physical_first, physical_last + (page_size -1));
  
  dprintf(STDERR_FILENO,"Loaded %ld page struct in memory DONE\n", elements);
  
  close (pfd);
  return 0;
}

int main (int argc, char * argv []) 
{
  // check arguments, here load the paddr file and the pids mappings
  unsigned long l;
  char * input_paddr = "paddr.stdout";
  
  page_size= getpagesize(); // we are assuming, as documents online that the bock size is a page
  
  //allocate memory first
  struct stat st;
  if (stat (input_paddr, &st) != 0) {
    dprintf(STDERR_FILENO, "ERROR cannot determine the size of %s\n", input_paddr);
    exit(1);
  }
  elems_array = (st.st_size / sizeof(paddr_desc));
  
  printf("file '%s' size %ld elements %ld(%ld) (elem size %d)\n",
	input_paddr, st.st_size, elems_array,
	(st.st_size % sizeof(paddr_desc)), (int)sizeof(paddr_desc));

  //re-build the page struct array
/*  pages_array = malloc (elems_array * sizeof(struct page));
  if (pages_array == 0) {
    dprintf(STDERR_FILENO, "ERROR malloc failure for size of %ld\n", elems_array * sizeof(struct page));
    exit(1);
  } */
  load_array(input_paddr, elems_array, pages_array); 
  
  return 0;
}