 
// Antonio Barbalace, SSRG Virginia Tech, 2015

// Main tool to run the fault injection analysis

#include <string.h>
#include <stdio.h>
#include <stdint.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/fcntl.h>
#include <sys/statfs.h>
#include <sys/stat.h>
 
#include "paddr.h"
#include "dproc.h"
#include "linux-config.h"
#include "memory-failure.h"
#include "page-types.h"

#define USE_PROCESS_DUMP

static int page_size;

static long elems_array = 0;
static struct page * pages_array = 0;

// is here the right place?
static int is_phys_mem_area(unsigned long);
int pfn_valid(unsigned long pfn) 
{
  if (!pages_array)
    return 0;
  return (pfn < elems_array) && // elems array is pfn_max
	 (pfn >= 0) &&
	 is_phys_mem_area(pfn);
	 //!((pages_array[pfn].mapping == 0) && (pages_array[pfn].flags == 0)); // totally wrong
}

// is here the right place?
struct page * pfn_to_page(unsigned long pfn)
{
  if (!pages_array)
    return 0;
  if (pfn >= elems_array)
    return 0;
  return &(pages_array[pfn]);
} 

// is here the right place?
unsigned long page_to_pfn(struct page * page)
{
  if (!pages_array)
    return 0;
  if (pages_array > page)
    return 0;
  return (page - pages_array);
} 

//////////////////////////////////////////////////////////////////////////////////
// Per-system
//////////////////////////////////////////////////////////////////////////////////

static unsigned long flags_to_flags(unsigned long k)
{
	unsigned long u = 0;
	
	u |= kpf_copy_bit(k, PG_locked, KPF_LOCKED); //u |= kpf_copy_bit(k, KPF_LOCKED,	PG_locked);

	u |= kpf_copy_bit(k, PG_slab, KPF_SLAB); //u |= kpf_copy_bit(k, KPF_SLAB, PG_slab);

	u |= kpf_copy_bit(k, PG_error, KPF_ERROR); //u |= kpf_copy_bit(k, KPF_ERROR,		PG_error);
	u |= kpf_copy_bit(k, PG_dirty, KPF_DIRTY); //u |= kpf_copy_bit(k, KPF_DIRTY,		PG_dirty);
	u |= kpf_copy_bit(k, PG_uptodate, KPF_UPTODATE); //u |= kpf_copy_bit(k, KPF_UPTODATE,	PG_uptodate);
	u |= kpf_copy_bit(k, PG_writeback, KPF_WRITEBACK); //u |= kpf_copy_bit(k, KPF_WRITEBACK,	PG_writeback);

	u |= kpf_copy_bit(k, PG_lru, KPF_LRU); //u |= kpf_copy_bit(k, KPF_LRU,		PG_lru);
	u |= kpf_copy_bit(k, PG_referenced, KPF_REFERENCED); //u |= kpf_copy_bit(k, KPF_REFERENCED,	PG_referenced);
	u |= kpf_copy_bit(k, PG_active, KPF_ACTIVE); //u |= kpf_copy_bit(k, KPF_ACTIVE,	PG_active);
	u |= kpf_copy_bit(k, PG_reclaim, KPF_RECLAIM); //u |= kpf_copy_bit(k, KPF_RECLAIM,	PG_reclaim);

	u |= kpf_copy_bit(k, PG_swapcache, KPF_SWAPCACHE); //u |= kpf_copy_bit(k, KPF_SWAPCACHE,	PG_swapcache);
	u |= kpf_copy_bit(k, PG_swapbacked, KPF_SWAPBACKED); //u |= kpf_copy_bit(k, KPF_SWAPBACKED,	PG_swapbacked);

	u |= kpf_copy_bit(k, PG_unevictable, KPF_UNEVICTABLE); //u |= kpf_copy_bit(k, KPF_UNEVICTABLE,	PG_unevictable);
	u |= kpf_copy_bit(k, PG_mlocked, KPF_MLOCKED); //u |= kpf_copy_bit(k, KPF_MLOCKED,	PG_mlocked);

#ifdef CONFIG_MEMORY_FAILURE
	u |= kpf_copy_bit(k, PG_hwpoison, KPF_HWPOISON); //u |= kpf_copy_bit(k, KPF_HWPOISON,	PG_hwpoison);
#else
	if (kpf_copy_bit(k, PG_hwpoison, KPF_HWPOISON)) {
		dprintf(STDERR_FILENO, "ERROR the tool has been compiled without CONFIG_MEMORY_FAILURE but this kernel has the option enabled. PLEASE RECOMPILE\n");
		exit(1);
	}
#endif

#ifdef CONFIG_ARCH_USES_PG_UNCACHED
	u |= kpf_copy_bit(k, PG_uncached, KPF_UNCACHED); //u |= kpf_copy_bit(k, KPF_UNCACHED,	PG_uncached);
#else 
	if (kpf_copy_bit(k, PG_uncached, KPF_UNCACHED)) {
		dprintf(STDERR_FILENO, "ERROR the tool has been compiled without CONFIG_ARCH_USES_PG_UNCACHED but this kernel has the option enabled. PLEASE RECOMPILE\n");
		exit(1);
	} 
#endif

	u |= kpf_copy_bit(k, PG_reserved, KPF_RESERVED); //u |= kpf_copy_bit(k, KPF_RESERVED,	PG_reserved);
	u |= kpf_copy_bit(k, PG_mappedtodisk, KPF_MAPPEDTODISK); //u |= kpf_copy_bit(k, KPF_MAPPEDTODISK,	PG_mappedtodisk);
	u |= kpf_copy_bit(k, PG_private, KPF_PRIVATE); //u |= kpf_copy_bit(k, KPF_PRIVATE,	PG_private);
	u |= kpf_copy_bit(k, PG_private_2, KPF_PRIVATE_2); //u |= kpf_copy_bit(k, KPF_PRIVATE_2,	PG_private_2);
	u |= kpf_copy_bit(k, PG_owner_priv_1, KPF_OWNER_PRIVATE); //u |= kpf_copy_bit(k, KPF_OWNER_PRIVATE,	PG_owner_priv_1);
	u |= kpf_copy_bit(k, PG_arch_1, KPF_ARCH); //u |= kpf_copy_bit(k, KPF_ARCH,		PG_arch_1);
	
	return u;
}

/* this is used while creating the page table therefore cannot tell me if the pfn is valid, for example ...
static int flags_page_mapcount(struct page * pp)
{
  if (!pfn_valid(page_to_pfn(pp)))
    return 0;
  
  if (PageSlab(pp)) // this follows fs/proc/page.c:kpagecount_read
    return 0;
  
  pp->_mapcount -= 1;
  return 0;
}
*/

/* flags_to_flags
 * 
 * This should run after fetching the page count, because it checks the page count 
 * for consistency
 */
struct page * compound_page_head = 0;
unsigned long compound_page_count = 0;
struct page * buddy_page_head = 0;
unsigned long buddy_page_count = 0;
extern void free_huge_page(struct page *page);
static int flags_to_page(struct page * pp, unsigned long flags, unsigned long count, unsigned long paddr)
{ 
	if ( flags & (1 << KPF_NOPAGE)) {
	  if ( (pp->mapping != 0) || (pp->_mapcount !=0) ) {
	    dprintf(STDERR_FILENO, "ERROR inconsistent, page should be NOPAGE but phys %ld _mapcount %ld\n",
		    pp->mapping, pp->_mapcount);
	    exit(1);
	  }
	  pp->flags = 0;
	  return 0;
	}
  
	if ( flags & (1 << KPF_MMAP) ) {
		if ( pp->_mapcount < 0 ) {
			dprintf(STDERR_FILENO, "ERROR flags_to_flags, mapcount SHOULD be >= 0, programming error?\n");
			exit(1);
		}
		__SetPageSlab(pp);
		set_page_count(pp, 1);
	}	
	if ( flags & (1 << KPF_ANON) )
		pp->mapping = (void*) (((unsigned long) pp->mapping) | PAGE_MAPPING_ANON);
	if ( flags & (1 << KPF_KSM) )
		pp->mapping = (void*) (((unsigned long) pp->mapping) | PAGE_MAPPING_KSM);

	if ( flags & (1 << KPF_COMPOUND_HEAD) ) {
		compound_page_head = pp;
		compound_page_count = 0;
		prep_compound_page(compound_page_head, compound_page_count);
	}
	if ( flags & (1 << KPF_COMPOUND_TAIL) ) {
		if ( compound_page_head == 0) {
		  dprintf(STDERR_FILENO, "ERROR Compound page TAIL without any head - incosistent state\n");
		  exit (1);
		}
		prep_compound_page(compound_page_head, ++compound_page_count);
	}

	if ( flags & (1 << KPF_HUGE) ) {
	  if ( !PageCompound(pp) ) {
	    dprintf(STDERR_FILENO, "ERROR cannot be huge if not compound\n");
	    exit(1);
	  }
	  struct page * ch = compound_head(pp);
	  set_compound_page_dtor(ch, free_huge_page); //refer to lwn.net/Articles/619514 for the differences between compound and Huge
	  //dprintf(STDERR_FILENO, "WARN, compound pages not fully supported\n");
	}

	if ( flags & (1 << KPF_BUDDY) ) {
	  pp->_mapcount = PAGE_BUDDY_MAPCOUNT_VALUE;
	  buddy_page_head = pp;
	  buddy_page_count = 1;
	  set_page_order(buddy_page_head, 0);
	}
	    
	pp->flags |= flags_to_flags(flags);

    if ( flags & (1 << KPF_BUDDY) ) { //another consistency check -- let's emulate the kernel here
//      struct page tmp = parray[l];
//      page_mapcount(&tmp);
//      if (tmp._mapcount != pdesc.count) {
      if (page_mapcount( pp) != count ) {
	dprintf(STDERR_FILENO, "ERROR this is inconsistent, kpagecount %ld expected %ld\n", pp->_mapcount, count);
	exit (1);
      }
    } 
    else {
      if (paddr != 0 && !PageSlab(pp))
	pp->_mapcount -= 1;
    }
	
	// we can maybe figure out a way to asses consistency here by finding a max alignment for each buddy ...
	if ( !(flags & (1 << KPF_BUDDY)) ) {
		if (buddy_page_head && (buddy_page_count > 0)) {
			// _count is 0 ...
			if (pp->flags == 0 && pp->_mapcount == -1) { //&& pp->mapping == 0
				buddy_page_count++;
				set_page_order(buddy_page_head, __fls(buddy_page_count));
			}
			else {
				buddy_page_head = 0;
				buddy_page_count = 0;
			
			}
		}
		else {
			buddy_page_head = 0;
			buddy_page_count = 0;
		}
	}

	return 0;
}

#define STRUCT_PAGE 0x40
 
// this contains the physical address of the page_table (areas)
struct page_table_area {
  struct page_table_area * next;
  unsigned long start;
  unsigned long end;
};
struct page_table_area * page_table = 0;
static int page_table_valid=0;

static int add_page_table_area(unsigned long start, unsigned long end)
{ 
  struct page_table_area * pta;
  struct page_table_area * _new = malloc(sizeof(struct page_table_area));
  if (!_new)
    return -1;
  
  _new->next = 0;
  _new->start = start;
  _new->end = end;
  
  if (page_table == 0) {
    page_table = _new;
    return 0;
  } 
  pta = page_table;
    
  while (pta->next != 0)
    pta = pta->next;
  pta->next = _new;
  
  return 0;
}
static int dump_page_table_area()
{
  struct page_table_area *pta=page_table;
  if (!page_table_valid) {
    dprintf(STDERR_FILENO, "ERROR Page Table area layout not valid.\n");
    exit(1);
  }
  while (pta != 0) {
    printf("%s: %lx - %lx\n", __func__, pta->start, pta->end);
    pta = pta->next;
  }
}
static int is_page_table_area(unsigned long addr)
{
  struct page_table_area *pta=page_table;
  while (pta != 0) {
    if ( (pta->start <= addr) && (addr <= (pta->end + STRUCT_PAGE -1)) )
      return 1;
    pta = pta->next;
  }
  return 0;
}

// this contains the physical address of the page_table (areas)
struct phys_mem_area {
  struct phys_mem_area * next;
  unsigned long start; //use PFN
  unsigned long end; // use PFN
};
struct phys_mem_area * phys_mem = 0;
static int phys_mem_valid=0;

static int add_phys_mem_area(unsigned long start, unsigned long end)
{ 
  struct phys_mem_area * pma;
  struct phys_mem_area * _new = malloc(sizeof(struct phys_mem_area));
  if (!_new)
    return -1;
  
  _new->next = 0;
  _new->start = start;
  _new->end = end;
  
  if (phys_mem == 0) {
    phys_mem = _new;
    return 0;
  } 
  pma = phys_mem;
    
  while (pma->next != 0)
    pma = pma->next;
  pma->next = _new;
  
  return 0;
}
static int dump_phys_mem_area()
{
  struct phys_mem_area *pma =phys_mem;
  while (pma != 0) {
    printf("%s: %lx - %lx\n", __func__, pma->start, pma->end);
    pma = pma->next;
  }
}
static int is_phys_mem_area(unsigned long pfn)
{
  struct phys_mem_area *pma =phys_mem;
  if (!phys_mem_valid) {
    dprintf(STDERR_FILENO, "ERROR Physical Memory area layout not valid.\n");
    exit(1);
  }
  while (pma != 0) {
    if ( (pma->start <= pfn) && (pfn <= pma->end) )
      return 1;
    pma = pma->next;
  }
  return 0;
}

static unsigned long group_first = (~0); // to identify page_table areas
static unsigned long group_last = (~0);
unsigned long physical_first = (~0); // to identify physical memory chuncks areas
unsigned long physical_last = (~0);

unsigned long load_array(char * input_file, long elements, struct page * parray)
{
  long l, bytes;
  long pfn_prev = -1, pfn_cur =0;
  int pfd;
  paddr_desc pdesc;
  
  pfd= open(input_file,O_RDONLY);
  if (pfd < 0) {
    perror(input_file);
    exit(1);
  }
  
  for (l=0; l<elements; l++) {
    bytes = read (pfd, (void *)&pdesc, sizeof(pdesc));
    if ( bytes < 0) {
      dprintf(STDERR_FILENO, "ERROR in reading the file %s\n", input_file);
      return -1;
    }
    if ( bytes % sizeof(pdesc) ) {
      dprintf(STDERR_FILENO, "ERROR in the alignement, read returned %ld (%ld) for %s\n", bytes, sizeof(pdesc), input_file);
      return -1;
    }
    
//    printf("%lx %lx 0x%08lx %ld 0x%08lx (%d)\n",
//	   pdesc.pfn, (pdesc.pfn /page_size), pdesc.phys, pdesc.count, pdesc.flags, bytes);

    //check for inconsistencies (I know this is evil ...)
    pfn_cur = pdesc.pfn / page_size;
    if ((pfn_prev +1) != pfn_cur) {// we strictly assume that pfn should be contiguous
      dprintf(STDERR_FILENO, "ERROR missing pfn, pfn are not contiguous pfn %ld prev %ld\n", pfn_cur, pfn_prev); 
      return -1;
    }
    if (pfn_cur != l) {
      dprintf(STDERR_FILENO, "ERROR something wrong, pfn are not contiguous pfn %ld counter %ld\n", pfn_cur, l); 
      return -1;
    }      
    
    memset ( &(parray[l]), 0, sizeof(struct page));
#ifndef USE_PROCESS_DUMP
    parray[l].mapping = (void*) pdesc.phys; // this was not a good idea therefore shouldn't be used, I leave it here for historical reasons
#endif
    parray[l]._mapcount = pdesc.count; // fs/proc/page.c:kpagecount_read if PageSlap returns 0 otherwise page_mapcount(ppage) i.e. ->_mapcount BUT mm.h fetches the value from page_mapcount()
    flags_to_page(&(parray[l]), pdesc.flags, pdesc.count, pdesc.phys);
    
#ifdef USE_PROCESS_DUMP
// page_table physical address areas extraction    
    if (pdesc.phys !=0 ) {
    if (group_first == (~0))
      group_first = group_last = pdesc.phys;
    else
      if ( (group_last + STRUCT_PAGE) == pdesc.phys )
        group_last = pdesc.phys;
      else {
        //printf("%lx - %lx\n", group_first, group_last);
	add_page_table_area(group_first, group_last);
        group_first = group_last = pdesc.phys;
      }
    }

// physical memory areas extraction
    if (pdesc.phys == 0 && (pdesc.flags & (1 << KPF_NOPAGE)) ) {
      if (physical_first !=(~0) && physical_last != (~0)) {
	//printf("physical %lx last %lx\n", physical_first, physical_last + (page_size -1));
	add_phys_mem_area(physical_first, physical_last); // + (page_size -1)); // uncomment if want to safe addresses
	physical_first = (~0);
	physical_last = (~0);
      }
    } else {
	if (physical_first == (~0))
	    physical_first = (pfn_cur); //*page_size); 
	else
	    physical_last = (pfn_cur); //*page_size); 
    }
#endif    
    
    pfn_prev=pfn_cur;
  }
  
#ifdef USE_PROCESS_DUMP  
  //printf("%lx - %lx\n", group_first, group_last);
  add_page_table_area(group_first, group_last);
    page_table_valid =1;
  add_phys_mem_area(physical_first, physical_last); // + (page_size -1));
    phys_mem_valid =1;
#endif  
  
  dprintf(STDERR_FILENO,"Loaded %ld page struct in memory (%ldB)\n", elements, elements*sizeof(struct page));
  
  close (pfd);
  return 0;
}

//////////////////////////////////////////////////////////////////////////////////
// Per-process
//////////////////////////////////////////////////////////////////////////////////  

/*
 * application's flags
 */
#define VMA_READ 0
#define VMA_WRITE 1
#define VMA_EXEC 2
#define VMA_SHARE 3

/*
 * pagemap kernel ABI bits
 */

#define PM_ENTRY_BYTES      sizeof(uint64_t)
#define PM_STATUS_BITS      3
#define PM_STATUS_OFFSET    (64 - PM_STATUS_BITS)
#define PM_STATUS_MASK      (((1LL << PM_STATUS_BITS) - 1) << PM_STATUS_OFFSET)
#define PM_STATUS(nr)       (((nr) << PM_STATUS_OFFSET) & PM_STATUS_MASK)

#define PM_PSHIFT_BITS      6
#define PM_PSHIFT_OFFSET    (PM_STATUS_OFFSET - PM_PSHIFT_BITS)
#define PM_PSHIFT_MASK      (((1LL << PM_PSHIFT_BITS) - 1) << PM_PSHIFT_OFFSET)
#define PM_PSHIFT(x)        (((u64) (x) << PM_PSHIFT_OFFSET) & PM_PSHIFT_MASK)

#define PM_PFRAME_MASK      ((1LL << PM_PSHIFT_OFFSET) - 1)
#define PM_PFRAME(x)        ((x) & PM_PFRAME_MASK)

#define PM_PRESENT          PM_STATUS(4LL)
#define PM_SWAP             PM_STATUS(2LL)
#define PM_FILE             PM_STATUS(1LL)
//#define PM_NOT_PRESENT      PM_PSHIFT(PAGE_SHIFT)
//#define PM_END_OF_BUFFER    1


static unsigned long pagemap_pfn(uint64_t val)
{
	unsigned long pfn;

	if (val & PM_PRESENT)
		pfn = PM_PFRAME(val);
	else
		pfn = 0;

	return pfn;
}

static unsigned long pagemap_swap(uint64_t val)
{
	unsigned long swap;
	
	if (val & PM_SWAP)
		swap = PM_PFRAME(val);
	else
		swap=0;
	
	return swap;
}

#define GET_PM_PSHIFT(x) (((unsigned long) (x) >> PM_PSHIFT_OFFSET) & ((1LL << PM_PSHIFT_BITS) - 1))

#define PRINT_READ(flags) ((flags & (1 << VMA_READ)) ? 'r' : '-')
#define PRINT_WRITE(flags) ((flags & (1 << VMA_WRITE)) ? 'w' : '-')
#define PRINT_EXEC(flags) ((flags & (1 << VMA_EXEC)) ? 'x' : '-')
#define PRINT_SHARE(flags) ((flags & (1 << VMA_SHARE)) ? 's' : 'p')
#define PRINT_FLAGS(flags) PRINT_READ(flags),PRINT_WRITE(flags),PRINT_EXEC(flags),PRINT_SHARE(flags)


static void list_insert_mask (struct address_space ** head, struct address_space * as, unsigned long mask)
{
  struct address_space * next = *head;
  as->next = next;
  *head = (struct address_space *)(((unsigned long)as & ~mask) | ((unsigned long)next & mask));
}
#define list_insert(a,b) list_insert_mask(a,b, (unsigned long)~0x0)

unsigned long load_desc(char * input_file, long elements, struct task_struct * tt)
{
  long l, bytes;
  unsigned long shift, pfn, swap;
  dproc_desc dproc;
  int pfd;
  int insert = 0;
 
  pfd= open(input_file,O_RDONLY);
  if (pfd < 0) {
    perror(input_file);
    exit(1);
  }
  
  for (l=0; l<elements; l++) {
    bytes = read (pfd, (void *)&dproc, sizeof(dproc));
    if ( bytes < 0) {
      dprintf(STDERR_FILENO, "load_desc ERROR in reading the file %s\n", input_file);
      return -1;
    }
    if ( bytes % sizeof(dproc) ) {
      dprintf(STDERR_FILENO, "load_desc ERROR in the alignement, read returned %ld (%ld) for %s\n", bytes, sizeof(dproc), input_file);
      return -1;
    }
    
    shift = GET_PM_PSHIFT(dproc.info); 
    	/*if (shift != 12)
	  dprintf(STDERR_FILENO, "load_desc WARN this is a bigger page! %d bit of SHIFT\n", shift);*/
    pfn = pagemap_pfn(dproc.info);
    swap = pagemap_swap(dproc.info);
    
    // There is a pfn ---------------------------------------------------------------------------------------
    if (pfn) {
      struct page * pp = pfn_to_page(pfn);
      
      if (shift != 12 && !PageCompound(pp) && !PageCompound(pp)) {
	 dprintf(STDERR_FILENO, "load_desc ERROR the page order is %d but is not Huge nor Compound\n", shift);
	 return -1;
      }
      
      struct address_space * mapping = malloc(sizeof(struct address_space));
      if (!mapping) {
	dprintf(STDERR_FILENO, "ERROR allocating descriptor\n");
	exit(1);
      }
      mapping->next = 0;
      mapping->task = tt;
      mapping->flags = dproc.flags;
      mapping->vaddr = dproc.addr;

// for the explanation of the following please refer to dump_process.c - not a huge problem but lack of kernel support      
#ifndef USE_PM_FILE
      if (PageAnon(pp)) {
	        list_insert_mask (&(pp->mapping), mapping, PAGE_MAPPING_FLAGS); insert++;
      }
      else {
	        list_insert (&(pp->mapping), mapping); insert++;
      }
#else      
      if (dproc.info & PM_FILE) { //file backed page
        //printf("%08lx pfn: %lx [%lx] %lx FILE %c%c%c%c\n", dproc.addr, pfn, dproc.info, shift, PRINT_FLAGS(dproc.flags));
        if ( PageAnon(pp) ) {
          dprintf(STDERR_FILENO, "load_desc ERROR the process mapping says file but physical mapping is anon\n"
                "[pfn: %lx dproc: %lx task: %s pid: %ld]\n", pfn, dproc.info, tt->comm, tt->pid);
          return -1;
        }
        list_insert (&(pp->mapping), mapping); insert++;
      }
      else { //anonimuous mapping page
        //printf("%08lx pfn: %lx [%lx] %lx ANON %c%c%c%c\n", dproc.addr, pfn, dproc.info, shift, PRINT_FLAGS(dproc.flags));
        if ( !PageAnon(pp) ) {
          dprintf(STDERR_FILENO, "load_desc ERROR the process mapping says anonimuous but physical mapping is not\n"
                "[pfn: %lx dproc: %lx task: %s pid: %ld]\n", pfn, dproc.info, tt->comm, tt->pid);
          return -1;
        }
        list_insert_mask (&(pp->mapping), mapping, PAGE_MAPPING_FLAGS); insert++;
      }
#endif      
    }
      //here there is no pfn therefore swap ... do some consistency checks -----------------------------
      else {
	continue; // should look at the source code in fs/proc/task_mmu.c but basically it is either present or swap
	/*if (swap)
	    //printf("%08lx swap: %lx [%lx] %lx %c%c%c%c\n", dproc.addr, swap, dproc.info, shift, PRINT_FLAGS(dproc.flags));
	    // I don't have any pfn -> therefore swapbacked?!
	else
	    //printf("%08lx none: %lx %lx %c%c%c%c\n", dproc.addr, dproc.info, shift, PRINT_FLAGS(dproc.flags));
*/
      }
   }
    dprintf(STDERR_FILENO,"dproc: Inserted %d elements\n", insert);
    return 0; 
}

//////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////

/* for the sake of writing decent code, this should be put in another source file, 
 * so that this file can be os dependent.
 */
int main (int argc, char * argv []) 
{
  // check arguments, here load the paddr file and the pids mappings
  unsigned long l; int ret;
  
  page_size= getpagesize(); // we are assuming, as documents online that the bock size is a page

//////////////////////////////////////////////////////////////////////////////////
// Load per-system data
//////////////////////////////////////////////////////////////////////////////////  
  char * input_paddr = "paddr.stdout";
  
  //allocate memory first
  struct stat st;
  if (stat (input_paddr, &st) != 0) {
    dprintf(STDERR_FILENO, "ERROR cannot determine the size of %s\n", input_paddr);
    exit(1);
  }
  elems_array= (st.st_size / sizeof(paddr_desc));
  
  printf("file '%s' size %ld elements %ld(%ld) (elem size %d)\n",
	input_paddr, st.st_size, elems_array,
	(st.st_size % sizeof(paddr_desc)), (int)sizeof(paddr_desc));

  //re-build the page struct array
  pages_array = malloc (elems_array * sizeof(struct page));
  if (pages_array == 0) {
    dprintf(STDERR_FILENO, "ERROR malloc failure for size of %ld\n", elems_array * sizeof(struct page));
    exit(1);
  }
  ret = load_array(input_paddr, elems_array, pages_array); // open file read it into the datastructure etc.
  if (ret)
        exit(1);
  
/*  dump_page_table_area();
  dump_phys_mem_area(); */ // for debug
    
//////////////////////////////////////////////////////////////////////////////////
// Load per-process data
//////////////////////////////////////////////////////////////////////////////////  
  char * input_dproc = "dproc.stdout";
  // TODO for each process in the system - handle input

  struct stat dpst;
  if (stat (input_dproc, &dpst) != 0) {
    dprintf(STDERR_FILENO, "ERROR cannot determine the size of %s\n", input_dproc);
    exit(1);
  }
  int descs = (dpst.st_size / sizeof(dproc_desc));

  printf("dproc file '%s' size %ld elements %ld(%ld) (elem size %d)\n",
        input_dproc, st.st_size, descs,
        (st.st_size % sizeof(dproc_desc)), (int)sizeof(dproc_desc));
  
  struct task_struct current; // TODO allocate an array of tasks
  strcpy(current.comm, "this_app"); current.pid=2006;
  ret = load_desc(input_dproc, descs, &current); // open file read it into the datastructure etc - some kind of linked lists
  if (ret)
        exit(1);
  
//////////////////////////////////////////////////////////////////////////////////
// Memory injection loop
//////////////////////////////////////////////////////////////////////////////////  
  
  // let's start simulate! inject error for each phys page
  for (l= 0; l <elems_array ; l++) {
    
    if ( is_page_table_area((l * page_size)) )
      printf("TOOL: %#lx: failed because is part of page table\n", l);
    
//    __memory_failure(l, 18, MF_COUNT_INCREASED); //following mm/hwpoison-inject.c: hwpoison_inject()
    int trapno = 18; // @trapno: Trap number reported in the signal to user space --- this is MCE_VECTOR
    __memory_failure(l, trapno, 0); //following mm/memory-failure.c: memory_failure()
  } 
  
  free(pages_array);
  
  return 1;
}
 